----------------------------------------------------------------------------
----------------------------------------------------------------------------
--                                                                        --
--     DSA 2 Exercise 1 (ex2)                                             --
--     Program Title : Fibonacci Algorithm Version 0.9                    --
--     Author : John Urquhart Ferguson                                    --
--     Group : L                                                          --
--     Matriculation Number : 0102662                                     --
--     Date Last Modified : 15th December 2003                            --
--                                                                        --
--     Description : This program takes in two integers from standard     --
--                   input and uses them as indices to calculate          --
--                   Fibonacci numbers from the lower to upper index.     --
--                                                                        --
--                   This program uses ListPackage for representing       --
--                   large integer values.                                --
--                                                                        --
--                                                                        --
--     State of Program: The program has worked correctly for all test    --
--                       data.                                            --
--                                                                        --
----------------------------------------------------------------------------
----------------------------------------------------------------------------

with Ada.Text_Io; use Ada.Text_Io;
with Ada.Integer_Text_IO; use Ada.Integer_Text_IO;
with Listpackage; use Listpackage;


procedure ex2 is
   
   -- Variables
   L    : Integer;
   U    : Integer;
   Swap : Integer;
   V    : Boolean;


   ----------------------------------------------------------------
   ----------------------------------------------------------------
   -- The Fib function calculates the value of the Fibonacci
   -- number with index n.
   function Fib (n : in integer) return Big is


      -------------------------------------------------------------
      -------------------------------------------------------------
      -- helpFib is used to speed up the Fib function by 
      -- preventing repeated calculations.
      function helpFib (n     : in integer;
                        k     : in integer;
                        fibk  : in Big    ;
                        fibk1 : in Big     ) return Big is
      begin
         if n = k then
            return fibk;
         else
            return helpFib(n, k+1, Add(fibk,fibk1), fibk);
         end if;
      end Helpfib;
      -------------------------------------------------------------
      -------------------------------------------------------------


      -- These Big values are required as input into the 
      -- helpFib function.
      One  : Big;
      Zero : Big;
      
   begin

      MakeOne(One);
      Makezero(Zero);
      
      -- For values less than 3, the Fib number is 1;
      if N < 3 then
         return One;
         
      -- For indices smaller than 21, use a recursive Fib function
      elsif N < 21 then
         return(Add(Fib(N-2), Fib(N-1)));  
         
      -- For indices larger than 20, use the helpFib function to
      -- speed things up.
      else
         return(helpFib(n, 1, One, Zero));
      end if;
      
   end Fib;
   ----------------------------------------------------------------
   ----------------------------------------------------------------


   ----------------------------------------------------------------
   ----------------------------------------------------------------
   -- The Verify procedure tests to see if the Fibonacci number
   -- calculated is accurate by trying to find it by a different 
   -- algorithm and then comparing it to the originally found 
   -- value.
   procedure Verify (N: in     Integer;
                     V:    out Boolean ) is
   test : big;
   begin
      if N /= 2 then
         Test := Fib(N);
         
         -- The Add function is taken from ListPackage
         Test := Add(Add(Test,Test),Test);
         
         -- IsEqual is taking from ListPackage
         if IsEqual(Test, (Add(Fib(N+2), Fib(N-2)))) then
            V := True;
         else
            V := False;
         end if;
      else
         V := True;
      end if;

   end Verify;
   ----------------------------------------------------------------
   ----------------------------------------------------------------


begin
   
   -- Get the indices to be used
   Get(L);
   Get(U);
   
   -- Check that numbers are in the correct range
   if (L < 1) or (U < 1) then
      Put ("Fibonacci Indices must be positive integers greater than zero.");
   else
      
      -- If indices are entered in the wrong order, reverse them
      if L > U then
         Swap := U;
         U := L;
         L := Swap;
      end if;
      
      -- Output the Fibonacci Numbers specified
      For I in L..U loop
         New_Line;
         Put("Fib(");
         Put(I,2);
         Put(") = ");
         
         -- WriteBig is from ListPackage
         Writebig(Fib(I));
         
         -- Verify that the calculated value is correct
         Verify(I,V);
         if V then
            Put ("  Valid");
         else
            Put ("  Invalid");
         end if;
         
      end loop;      
   end if;
   
end ex2;